/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Client virtual memory Management
*  SCOPE :              Client virtual memory Management
*  FILE :               virtualmemory.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma implementation
#endif

/* IMPORTS *******************************************************************/

#ifndef __fnctl_H
#define __fnctl_H
#include <fcntl.h>
#endif

#ifndef __unistd_H
#define __unistd_H
#include <unistd.h>
#endif

#ifndef __signal_H
#define __signal_H
#include <sys/signal.h>
#endif

#ifndef __virtualmemory_HXX
#include "virtualmemory.hxx"
#endif

#ifndef __hem_CC
#include "hem.cc"
#endif

#ifndef __var_H
#include "var.h"
#endif

#ifndef __define_H
#include "define.h"
#endif

#ifndef __yoofunc_HXX
#include "yoofunc.hxx"
#endif

#ifndef __page_HXX
#include "page.hxx"
#endif

/*****************************************************************************/


/*= EXTERNAL FUNCTION =======================================================*/
VirtualDBMemory::VirtualDBMemory(int size)
: _size(size), _allocator(size), _last(0)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (_size > YOO_MAX_VM_SIZE) {
    yoo_error(YOO_ERR_MEM,
	      "Impossible to create Virtual memory. Size to large") ;
    yoo_exit(1) ;
  }
  _allocator.zero() ;
  _name = tempnam(YOO_VM_DIR, "yoovm") ;
  SYSCALLI(this->_fd = open(_name, O_CREAT|O_RDWR,0666)) ;
  SYSCALLI(ftruncate(this->_fd, 2*_size*PAGE_SIZE)) ;
  this->_inMemoryTable = new PageHdlMem*[_size] ;
  memset((char*)_inMemoryTable, 0, _size*sizeof(PageHdlMem*)) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
VirtualDBMemory::~VirtualDBMemory()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  IterPageSet iter(_inVMPagePID) ;
  for (; iter.more(); iter.next()) {
    delete ((PageHdlMem*)iter.get()) ;
  }

  delete this->_inMemoryTable ;
  SYSCALLI(unlink(_name)) ;
  delete _name ;
}


/*= EXTERNAL FUNCTION =======================================================*/
PageHdlMem* VirtualDBMemory::allocation(PID pid, int size, Lock l)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
#ifndef __GNUG__
  PageHdlMem** pd_t = new PageHdlMem*[size] ;
#else  
  PageHdlMem* pd_t[size] ;
#endif

  int num = this->findAllocation() ;

  Page* p;
  if (size == 1) {
    p = &this->_currmap[num] ;
  } else {
    // Use to reserve virtual memory in the process
    p = (Page*)yoo_mmap(0, PAGE_SIZE*size, PROT_NONE, MAP_SHARED, _fd,
			num*PAGE_SIZE) ;
  }

  PageHdlMem *first, *next ;
  for (int i = 0; i < size; i++) {
    first = i == 0 ? NULL : pd_t[0] ;
    pd_t[i] = new PageHdlMem(&p[i], pid, l, first, NULL) ;
    _inVMPagePID += pd_t[i] ;
    _inVMPageAddr += pd_t[i] ;
    pid += 1 ;
  }

  for (i = 0; i < size; i++) {
    next = (i+1) == size ? NULL : pd_t[i+1] ;
    pd_t[i]->setnext(next) ;
  }

  pd_t[0]->inMemory() = num ;
  _inMemoryTable[num] = pd_t[0] ;
  this->physicalAllocation(pd_t[0]) ;
  
#ifndef __GNUG__
  delete [] pd_t ;
#endif

  return pd_t[0] ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int VirtualDBMemory::findAllocation()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (_last == 0) {
    this->_currmap = (Page*)yoo_mmap(0,PAGE_SIZE*YOO_VM_SIZE,PROT_NONE,
				 MAP_SHARED,_fd,0) ;
//    madvise((char*)this->_currmap, PAGE_SIZE*YOO_VM_SIZE, MADV_RANDOM) ;
  }
  int num = this->_allocator.look(_last, 1) ;
  if (num == -1) {
    this->_currmap = (Page*)yoo_mmap(0,PAGE_SIZE*YOO_VM_SIZE,PROT_NONE,
				     MAP_SHARED,_fd,0) ;
    _last = 0 ; num = this->_allocator.look(_last, 1) ; }
  
  if (num == -1) {
    yoo_error(YOO_ERR_MEM, "Memory full") ;
    yoo_exit(YOO_ERR_MEM) ;
  }

  if (_inMemoryTable[num] != NULL) {
    _inMemoryTable[num]->eject() ;
  }

  _last = (num+1) >= _size ? 0 : num + 1 ;

  return num ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void VirtualDBMemory::physicalAllocation(PageHdlMem* pd)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (pd->inMemory() != -1) {
    SYSCALLI(mprotect((caddr_t)pd->page(), PAGE_SIZE, PROT_READ|PROT_WRITE)) ;
  } else {
    int num = this->findAllocation() ;
    _inMemoryTable[num] = pd ;
    pd->inMemory() = num ;

    // Undo reservation mapping to do real mapping
    Page* res = (Page*)mmap((caddr_t)pd->page(),PAGE_SIZE,PROT_READ|PROT_WRITE,
			    MAP_FIXED|MAP_SHARED, _fd, num*PAGE_SIZE) ;
    if (res != pd->page()) {
      yoo_error(YOO_ERR_MEM, "Virtual Memory full") ;
      yoo_exit(YOO_ERR_MEM) ;
    }
  }

  pd->load() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void VirtualDBMemory::refresh(PageHdlMem* pd)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  _inVMPageAddr += pd ;
  _inVMPagePID += pd ;
  pd->refresh() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void VirtualDBMemory::eject(PageHdlMem* pd)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  _inVMPageAddr -= pd ;
  _inVMPagePID -= pd->pid() ;
  pd->eject() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void VirtualDBMemory::reset()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  _allocator.zero() ;
  IterPageSet iter(_inVMPagePID) ;
  for (; iter.more(); iter.next()) {
    if (((PageHdlMem*)iter.get())->iskept() == FALSE) {
      if (((PageHdlMem*)iter.get())->inMemory() != -1) {
	_inMemoryTable[((PageHdlMem*)iter.get())->inMemory()] = NULL ;
      }
      delete ((PageHdlMem*)iter.get()) ;
    }
  }
  _inVMPagePID.clear() ;
  _inVMPageAddr.clear() ;
}


#if defined sun && !defined __svr4__
void VirtualDBMemory::segvhandler(int,int code,struct sigcontext*, char* addr) 
{
#elif defined(__alpha__) || defined(__hpux__)
void VirtualDBMemory::segvhandler(int,int code,struct sigcontext* ctx) 
{
#ifdef __hpux__
  char* addr = (char*)ctx->sc_sl.sl_ss.ss_cr21 ;
#else 
  // Find page descriptor of the faulted address.
  char* addr = (char*)ctx->sc_traparg_a0 ;
#endif
#elif defined sun && defined __svr4__
void VirtualDBMemory::segvhandler(int, siginfo_t* ctx, ucontext_t* uc) 
{
  // Find page descriptor of the faulted address.
  char* addr = (char*)ctx->si_addr ;

#endif
  
  PageHdlMem* pd = yoo_getpage(addr) ;

  // Not PROTECTION error, address outside of virtual memory or deleted page
  // in such case, it is a real segmentation violation.
  if ( (pd==NULL) || (pd->isdeleted()) ) {
    yoo_error(YOO_ERR_SEGV, "Segmentation violation") ;
    yoo_exit(1) ;
  }

  if (pd->inMemory() == -1) {
    YOO_VirtualMemory->physicalAllocation(pd) ;
    if (YOO_RECV_PHASE) {
      SYSCALLI(mprotect((caddr_t)pd->page(), PAGE_SIZE, PROT_READ|PROT_WRITE));
    }
  } else {
    if (YOO_RECV_PHASE) {
      SYSCALLI(mprotect((caddr_t)pd->page(), PAGE_SIZE, PROT_READ|PROT_WRITE));
    } else {
      if (pd->getLock() >= YOO_EXCL) {
	yoo_error(YOO_ERR_BUS, "Bus error") ;
	yoo_exit(1) ;
      } else {	
	pd->xlock() ;
      }
    }  
  }
}


